import numpy as np
import matplotlib.pyplot as plt
import random as rand
from . import trial
from . import session
import copy



def shuffle(Base):
    """
    Randomly Shuffles Trials in Epoch and Sessions
    """

    base = Base
    id = 0
    if isinstance(base, session.Session):
        id = 0
    elif isinstance(base, session.Epochs):
        id = 1
    else:
        raise ValueError("Only instances of the base class may be shuffled, Epochs or Session")
        return

    num_trial = len(base.data)
    instances = base.data
    conditions = base.conditions
    rand_idx = []

    while len(rand_idx) != num_trial:
        new = rand.randrange(0,num_trial)
        if new in rand_idx:
            continue
        else:
            rand_idx.append(new)

    new_instances = []
    new_conditions = []

    for i in rand_idx:
        instance = instances[i]
        new_instances.append(instance)
        cond = conditions[i]
        new_conditions.append(cond)

    if id == 0:
        return Session(np.array(new_instances), np.array(new_conditions))
    else:
        return Epochs(np.array(new_instances), np.array(new_conditions))



class Base:
    """
    Base Class for Session and Epoch
    """
    def __init__(self, data, conditions=None):
        self.data = data
        if conditions is not None:
            self.conditions = conditions
        else:
            self.conditions = np.zeros(len(self.data))



    def summary(self, verbose = False):
        """
        summary: Dictionary of summary statistic for pupil size
            > mean: pupil mean
            > variance: pupil variance
            > stdev: pupil standard deviation
            > length: pupil size
            > min: pupil minimum
            > maximum: pupil maximum
        """

        summary = {}

        # Pupil Data
        pup_data = []
        for i in self.data:
            for j in i.pupil_size:
                pup_data.append(j)

        # Statistics and Shape of pupil_size across all session
        pupil_mean = np.nanmean(pup_data)
        summary['mean'] = pupil_mean
        pupil_variance = np.nanvar(pup_data)
        summary['variance'] = pupil_variance
        pupil_stddev = np.nanstd(pup_data)
        summary['stdev'] = pupil_stddev
        pupil_size = len(pup_data)
        summary['length'] = pupil_size
        pupil_min = np.nanmin(pup_data)
        summary['min'] = pupil_min
        pupil_max = np.nanmax(pup_data)
        summary['max'] = pupil_max

        if verbose:
            print("Session Pupil Mean: ", pupil_mean, '\n'
                    "Session Pupil Variance: ", pupil_variance, '\n'
                    "Session Pupil Standard Deviation: ", pupil_stddev, '\n'
                    "Session Pupil Data Length: ", pupil_size, '\n'
                    "Session Minimum Pupil Size: ", pupil_min, '\n'
                    "Session Maximum Pupil Size: ", pupil_max)

        return summary



    def get_trial(self, trial_num):
        """
        Return Specified Trial
        """

        trial = self.data[trial_num]
        return trial



    def get_pupil(self):
        """
        Returns Numpy Array Of Trials Containing Pupil Sizes
        """

        tmp_ls = []
        for i in self.data:
            tmp_ls.append(i.pupil_size)
        return np.array(tmp_ls)



    def get_movement(self):
        """
        Returns Numpy Array Of Trials Containing Movements
            > index 0: x-coordinates
            > index 1: y-coordinates
        """

        move_list = [[],[]]
        for i in self.data:
            move_list[0].append(i.movement_X)
            move_list[1].append(i.movement_y)
        return np.array(move_list)

    def get_timestamps(self):
        """
        Returns Numpy Array Of Trials Containing Timestamps
        """

        tmp_ls = []
        for i in self.data:
            tmp_ls.append(i.timestamps)
        return np.array(tmp_ls)



    def get_fixation(self):
        """
        Returns Numpy Array Of Trials Containing
         x And y Fixations Coordinates
         ie. Trial 1 x-coordinate
         > fixation[0][0][0]
         ie. Trial 2 y-coordinate
         > fixation[1][0][1]
        """

        tmp_ls = np.array(self.get_timestamps())
        movement_list = self.get_movement()
        fix = [[] for _ in range(tmp_ls.shape[0])]
        fix_list = [[] for _ in range(tmp_ls.shape[0])]
        RL = ''

        count = 0
        if(len(self.data[0].event_L) > 0):
            RL = 'L'
        elif(len(self.data[0].event_R ) > 0):
            RL = 'R'
        else:
            print('NO FIXATION FOUND')
            pass
        if(RL == 'L'):
            for i in tmp_ls:
                for j in self.data:
                    if(j.event_L[1] in i):
                        fix[count].append([j.event_L[0],j.event_L[1]])
                count += 1
        elif(RL == 'R'):
            for i in tmp_ls:
                index = 0
                for h in self.data:
                    for j in h.event_R:
                        if(j[1] in i):
                            fix[count].append([j[0],j[1]])
                count += 1

        for index in range(len(tmp_ls)):
            movements = [[],[]]
            in_fix = False
            for fix_index in fix[index]:
                for time_index in range(len(tmp_ls[index])):
                    if(fix_index[0] == tmp_ls[index][time_index]):
                        in_fix = True
                        # print(1)
                        movements[0].append(movement_list[0][index][time_index])
                        # print(movement_list[0][index][time_index])
                        movements[1].append(movement_list[1][index][time_index])
                    elif(fix_index[1] == tmp_ls[index][time_index]):
                        in_fix = False
                        movements[0].append(movement_list[0][index][time_index])
                        movements[1].append(movement_list[1][index][time_index])

                    if(in_fix):
                        movements[0].append(movement_list[0][index][time_index])
                        movements[1].append(movement_list[1][index][time_index])
            fix_list[index].append(movements)

        return np.array(fix_list)



    def select(self, indexes):
        """
        Selects Specified Trial By Index Value
        """

        new = copy.deepcopy(self)
        new_data = []
        new_cond = []
        for i in indexes:
            new_data.append(self.data[i])
            new_cond.append(self.conditions[i])
        new.data = new_data
        new.conditions = new_cond

        return new



    def plot_pupil_size(self, trial_num):
        """
        Plots pupil size of a specified trial over a period of time
        """

        index = trial_num - 1
        trial = self.get_trial(trial_num)
        pupil = trial.pupil_size
        plt.plot(range(len(pupil)), pupil)
        plt.title('Pupil_size: Trial %1.f' % trial_num)
        plt.show()



    def plot_tragectory(self, trial_num):
        """
        Plots Eye Movement Tragectory for Specified Trial
        """

        index = trial_num - 1
        m = self.get_movement()
        plt.xlim(0,1920)
        plt.ylim(0,1080)
        plt.plot(m[1][index], m[0][index])
        plt.title('Movement: Trial %1.f' % trial_num)
        plt.xlabel('Horizontal Eye Movement')
        plt.ylabel('Vertical Eye Movement')
        plt.show()



    def plot_xy(self, trial_num):
        """
        Plots x-coordinates and y-coordinate over a period of time
        """

        index = trial_num - 1
        m = self.get_movement()
        plt.plot(range(len(m[0][index])), m[0][index], label = 'x-axis movement')
        plt.plot(range(len(m[0][index])), m[1][index], label = 'y-axis movement')
        plt.legend()
        plt.title('x-axis vs y-axis Movement: Trial %1.f' % trial_num)
        plt.show()



class Session(Base):
    """
    Subclass of Base that hold whole trials
        > Reserved for later development for trials
    """

    def __init__(self, trials, conditions=None):
        Base.__init__(self, trials, conditions)



class Epochs(Base):
    """
    Subclass of Base that hold time locking epochs
        > Reserved for later development for Epochs
    """

    def __init__(self, epochs, conditions=None):
        Base.__init__(self, epochs, conditions)



import numpy as np



class Trial:
    """
    The Trial Class For Individual trials Within The 'data' Data Structure

    timestamps: list of timestamps
    pupil_size: list of pupil sizes
    movement_X: list of movements in x-axis
    movement_y: list of movements in y-axis
    event_L: list of fixation events for the left eye
    event_R: list of fixation events for the right eye
    """

    def __init__(self, timestamps, pupil_size, movement_X = [], movement_y = [], event_L = [], event_R = []):
        self.timestamps = timestamps
        self.pupil_size = pupil_size
        self.movement_X = movement_X
        self.movement_y = movement_y
        self.event_L = event_L
        self.event_R = event_R

    def summary(self, verbose = False):
        """
        summary: Dictionary of summary statistic for pupil size
            > mean: pupil mean
            > variance: pupil variance
            > stdev: pupil standard deviation
            > length: pupil size
            > min: pupil minimum
            > maximum: pupil maximum
        """
        summary = {}

        # Pupil Data
        pup_data = self.pupil_size

        # Statistics and Shape of pupil_size across all session
        pupil_mean = np.nanmean(pup_data)
        summary['mean'] = pupil_mean
        pupil_variance = np.nanvar(pup_data)
        summary['variance'] = pupil_variance
        pupil_stddev = np.nanstd(pup_data)
        summary['stdev'] = pupil_stddev
        pupil_size = len(pup_data)
        summary['length'] = pupil_size
        pupil_min = np.nanmin(pup_data)
        summary['min'] = pupil_min
        pupil_max = np.nanmax(pup_data)
        summary['max'] = pupil_max

        if verbose:
            print("Session Pupil Mean: ", pupil_mean, '\n'
                    "Session Pupil Variance: ", pupil_variance, '\n'
                    "Session Pupil Standard Deviation: ", pupil_stddev, '\n'
                    "Session Pupil Data Length: ", pupil_size, '\n'
                    "Session Minimum Pupil Size: ", pupil_min, '\n'
                    "Session Maximum Pupil Size: ", pupil_max)

        return summary



import matplotlib.pyplot as plt
import matplotlib
matplotlib.use("TkAgg")
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk
from matplotlib.figure import Figure
import numpy as np

import tkinter as tk
from tkinter import ttk
from tkinter import messagebox
from . import data


#creates initial window for visualization
class Visualization(tk.Tk):

    def __init__(self, data, *args, **kwargs):

        tk.Tk.__init__(self, *args, **kwargs)
        container = tk.Frame(self)

        container.pack(side="top", fill="both", expand = True)

        container.grid_rowconfigure(0, weight=1)
        container.grid_columnconfigure(0, weight=1)

        self.data = data

        #data has trials values of pupil size and timestamps
        self.pup_size = []
        self.timestmp = []
        self.index = 0
        #0 = bad and 1 = good
        self.curr_quailty = 0
        self.pupil_trials = self.data.get_values()
        self.timestamp_trials = self.data.get_timestamps()
        self.quality_list = np.ones(len(self.pupil_trials)) # change to all 1

        #initalize timestamps in milliseconds(0-end)
        for t in range(len(self.timestamp_trials)):
            self.timestamp_trials[t] -= self.timestamp_trials[t][0]

        good_button = ttk.Button(self,
                            text="Good",
                            command=lambda: self.good_button_function())
        good_button.pack(side = tk.TOP)

        bad_button = ttk.Button(self,
                            text="Bad",
                            command=lambda: self.bad_button_function())
        bad_button.pack(side = tk.TOP)

        next_button = ttk.Button(self,
                            text="Next",
                            command=lambda: self.next_graph())
        next_button.pack(side = tk.TOP)

        previous_button = ttk.Button(self,
                            text="Previous",
                            command=lambda: self.previous_graph())
        previous_button.pack(side = tk.TOP)

        self.my_text = ttk.Label(self,
         text= f"Graph: {self.index +1}     Quality: {self.quality_list[self.index]}")
        self.my_text.pack(side = tk.TOP)

        self.f = Figure(figsize=(5,5), dpi=100)
        self.a = self.f.add_subplot(111)
        self.a.plot(self.timestamp_trials[self.index],
                    self.pupil_trials[self.index])
        self.a.set_title('Pupil Size Trial {}'.format(str(self.index+1)))
        self.a.set_xlabel('Time (ms)')
        self.a.set_ylabel('Pupil Size')

        self.canvas = FigureCanvasTkAgg(self.f, self)
        self.canvas.draw()
        self.toolbar = NavigationToolbar2Tk(self.canvas, self)
        self.toolbar.update()
        self.canvas._tkcanvas.pack(side=tk.BOTTOM,
                                   fill=tk.BOTH,
                                   expand=True)
        self.canvas.get_tk_widget().pack(side=tk.BOTTOM,
                                         fill=tk.BOTH,
                                         expand=True)



    #gets next graph by getting adding 1 to the current position (index)
    def next_graph(self):
        # get pupil size data from index
        self.a.clear()
        self.canvas.get_tk_widget().destroy()
        self.index += 1

        if(self.index >= len(self.pupil_trials)):
            messagebox.askquestion('Finished',
                        'This is the last graph\nDo you want to exit?')
            if finished == 'yes':
                    self.quit()
            self.index -= 1
            pass
        self.f = Figure(figsize=(5,5), dpi=100)
        self.a = self.f.add_subplot(111)
        self.a.plot(self.timestamp_trials[self.index],
                    self.pupil_trials[self.index])
        self.a.set_title('Pupil Size Trial {}'.format(str(self.index+1)))
        self.a.set_xlabel('Time (ms)')
        self.a.set_ylabel('Pupil Size')

        self.canvas = FigureCanvasTkAgg(self.f, self)
        self.canvas.draw()
        self.canvas.get_tk_widget().pack(side=tk.BOTTOM,
                                         fill=tk.BOTH,
                                         expand=True)

        self.toolbar.update()
        self.canvas._tkcanvas.pack(side=tk.BOTTOM,
                                   fill=tk.BOTH,
                                   expand=True)
        self.my_text.pack_forget()
        self.my_text = ttk.Label(self,
                                text= f"Graph: {self.index +1}     {self.quality_list[self.index]}")
        self.my_text.pack(side = tk.TOP)

    #gets previous graph by subtracting 1 to the current position (index)
    def previous_graph(self):
        # get pupil size data from index
        self.a.clear()
        self.canvas.get_tk_widget().destroy()
        self.index -= 1
        if(self.index < 0):
            messagebox.showinfo('Out of Index',
                            'You are at the begining of the List')
            self.index = 0
            pass
        self.f = Figure(figsize=(5,5), dpi=100)
        self.a = self.f.add_subplot(111)
        self.a.plot(self.timestamp_trials[self.index],
                    self.pupil_trials[self.index])
        self.a.set_title('Pupil Size Trial {}'.format(str(self.index+1)))
        self.a.set_xlabel('Time (ms)')
        self.a.set_ylabel('Pupil Size')

        self.canvas = FigureCanvasTkAgg(self.f, self)
        self.canvas.draw()
        self.canvas.get_tk_widget().pack(side=tk.BOTTOM,
                                         fill=tk.BOTH,
                                         expand=True)
        self.toolbar.update()
        self.canvas._tkcanvas.pack(side=tk.BOTTOM,
                                   fill=tk.BOTH,
                                   expand=True)

        self.my_text.pack_forget()
        self.my_text = ttk.Label(self,
                                 text= f"Graph: {self.index +1}     {self.quality_list[self.index]}")
        self.my_text.pack(side = tk.TOP)

    #if current graph is good, change value to 1
    def good_button_function(self):
        self.quality_list[self.index] = 1

        self.my_text.pack_forget()
        self.my_text = ttk.Label(self,
                                 text= f"Graph: {self.index +1}     {self.quality_list[self.index]}")
        self.my_text.pack(side = tk.TOP)

    #if current grpah is bad, change value to 0
    def bad_button_function(self):
        self.quality_list[self.index] = 0

        self.my_text.pack_forget()
        self.my_text = ttk.Label(self,
                                 text= f"Graph: {self.index +1}     {self.quality_list[self.index]}")
        self.my_text.pack(side = tk.TOP)



# object unit testing
from abra import data
from abra import trial
from abra import session
import numpy as np


default_obj = data.read("test/asc/1211NE1.asc")
udef_obj = data.read("test/asc/22205.asc", mode = 'u')


def test_trial_object():
    # Test for default case:
    sess = default_obj.create_session()
    for t in sess.data:
        assert isinstance(t, trial.Trial)
        assert t.timestamps.ndim == 1
        assert t.pupil_size.ndim == 1
        assert isinstance(t.timestamps, np.ndarray)
        assert isinstance(t.pupil_size, np.ndarray)

        sum = t.summary()
        assert isinstance(sum['mean'], float)
        assert isinstance(sum['variance'], float)
        assert isinstance(sum['stdev'], float)
        assert isinstance(sum['length'], int)
        assert isinstance(sum['min'], float)
        assert isinstance(sum['max'], float)

    # Test for user define case:
    sess = udef_obj.create_session()
    for t in sess.data:
        assert isinstance(t, trial.Trial)
        assert t.timestamps.ndim == 1
        assert t.pupil_size.ndim == 1
        assert isinstance(t.timestamps, np.ndarray)
        assert isinstance(t.pupil_size, np.ndarray)

        sum = t.summary()
        assert isinstance(sum['mean'], float)
        assert isinstance(sum['variance'], float)
        assert isinstance(sum['stdev'], float)
        assert isinstance(sum['length'], int)
        assert isinstance(sum['min'], float)
        assert isinstance(sum['max'], float)

def test_session_object():
    sess = default_obj.create_session()
    assert isinstance(sess, session.Session)
    #assert isinstance(sess.data, list)
    assert isinstance(sess.data[0], trial.Trial)
    assert isinstance(sess.conditions, np.ndarray)
    assert len(sess.data) == len(sess.conditions)

    sum = sess.summary()
    assert isinstance(sum['mean'], float)
    assert isinstance(sum['variance'], float)
    assert isinstance(sum['stdev'], float)
    assert isinstance(sum['length'], int)
    assert isinstance(sum['min'], float)
    assert isinstance(sum['max'], float)

    index = [0,1,5,7,8,4,20,15,14,22,30]
    selected = sess.select(index)
    assert isinstance(selected, session.Session)
    assert len(selected.data) == len(index)

    sess = udef_obj.create_session()
    assert isinstance(sess, session.Session)
    #assert isinstance(sess.data, list)
    assert isinstance(sess.data[0], trial.Trial)
    assert isinstance(sess.conditions, np.ndarray)
    assert len(sess.data) == len(sess.conditions)

    sum = sess.summary()
    assert isinstance(sum['mean'], float)
    assert isinstance(sum['variance'], float)
    assert isinstance(sum['stdev'], float)
    assert isinstance(sum['length'], int)
    assert isinstance(sum['min'], float)
    assert isinstance(sum['max'], float)

    index = [0,1,5,7,8,4,20,15,14,22,30]
    selected = sess.select(index)
    assert isinstance(selected, session.Session)
    assert len(selected.data) == len(index)

def test_epochs_object():
    cleaned_default = data.pupil_size_remove_eye_blinks(default_obj, buffer=50)
    event_timestamps = np.array(cleaned_default.trial_markers['start']) + 1000  # Buffer for baslining
    test_epochs = cleaned_default.create_epochs(event_timestamps,
                                                conditions=None,
                                                pre_event=200,
                                                post_event=200,
                                                pupil_baseline=[-200,-100])
    sum = test_epochs.summary()
    assert isinstance(sum['mean'], float)
    assert isinstance(sum['variance'], float)
    assert isinstance(sum['stdev'], float)
    assert isinstance(sum['length'], int)
    assert isinstance(sum['min'], float)
    assert isinstance(sum['max'], float)

    assert test_epochs.get_pupil().shape[1] == 200

    index = [0,1,5,7,8,4,20,15,14,22,30]
    selected = test_epochs.select(index)

    assert isinstance(selected, session.Epochs)
    assert len(selected.data) == len(index)

    cleaned_udef = data.pupil_size_remove_eye_blinks(udef_obj, buffer=50)
    event_timestamps = np.array(cleaned_udef.trial_markers['start']) + 1000  # Buffer for baslining
    test_epochs = cleaned_udef.create_epochs(event_timestamps,
                                                conditions=None,
                                                pre_event=200,
                                                post_event=200,
                                                pupil_baseline=[-200,-100])
    sum = test_epochs.summary()
    assert isinstance(sum['mean'], float)
    assert isinstance(sum['variance'], float)
    assert isinstance(sum['stdev'], float)
    assert isinstance(sum['length'], int)
    assert isinstance(sum['min'], float)
    assert isinstance(sum['max'], float)

    assert test_epochs.get_pupil().shape[1] == 200

    index = [0,1,5,7,8,4,20,15,14,22,30]
    selected = test_epochs.select(index)

    assert isinstance(selected, session.Epochs)
    assert len(selected.data) == len(index)

def test_data_object():
    #Test for missing values in messages, event, and trial markers
    assert None not in udef_obj.timestamps
    assert None not in udef_obj.pupil_size
    for mes in udef_obj.messages:
        assert len(udef_obj.messages[mes]) > 0
    for evnt in udef_obj.events:
        assert len(udef_obj.events[evnt]) > 0
    for ts in udef_obj.trial_markers["start"]:
        assert ts is not None
    for te in udef_obj.trial_markers["end"]:
        assert ts is not None

    assert None not in default_obj.timestamps
    assert None not in default_obj.pupil_size
    for mes in default_obj.messages:
        assert len(default_obj.messages[mes]) > 0
    for evnt in default_obj.events:
        assert len(default_obj.events[evnt]) > 0
    for ts in default_obj.trial_markers["start"]:
        assert ts is not None
    for te in default_obj.trial_markers["end"]:
        assert te is not None

    #Test to make sure variables have the same length
    assert default_obj.timestamps.size == default_obj.pupil_size.size
    assert default_obj.timestamps.size == default_obj.movement[0].size
    assert default_obj.timestamps.size == default_obj.movement[1].size
    assert default_obj.movement[0].size == default_obj.movement[1].size


    assert udef_obj.timestamps.size == udef_obj.pupil_size.size
    assert udef_obj.timestamps.size == udef_obj.movement[0].size
    assert udef_obj.timestamps.size == udef_obj.movement[1].size
    assert udef_obj.movement[0].size == udef_obj.movement[1].size

    #Test dimensions
    assert default_obj.timestamps.ndim == 1
    assert default_obj.pupil_size.ndim == 1
    assert default_obj.movement.ndim == 2

    assert udef_obj.timestamps.ndim == 1
    assert udef_obj.pupil_size.ndim == 1
    assert udef_obj.movement.ndim == 2

    #Test data member type
    assert isinstance(default_obj.timestamps, np.ndarray)
    assert isinstance(default_obj.pupil_size, np.ndarray)
    assert isinstance(default_obj.movement, np.ndarray)
    assert isinstance(default_obj.sample_rate, int)
    assert isinstance(default_obj.calibration, dict)
    assert isinstance(default_obj.messages, dict)
    assert isinstance(default_obj.events, dict)

    assert isinstance(udef_obj.timestamps, np.ndarray)
    assert isinstance(udef_obj.pupil_size, np.ndarray)
    assert isinstance(udef_obj.movement, np.ndarray)
    assert isinstance(udef_obj.sample_rate, int)
    assert isinstance(udef_obj.calibration, dict)
    assert isinstance(udef_obj.messages, dict)
    assert isinstance(udef_obj.events, dict)



# method unit testing
from abra import data
from abra import trial
from abra import session
import numpy as np

default_obj = data.read('test/asc/1211NE1.asc')
udef_obj = data.read('test/asc/22205.asc', mode = 'u')

def test_read_output():
    assert isinstance(default_obj, data.Data)
    assert isinstance(udef_obj, data.Data)


def test_remove_eye_blinks():
    processed = data.pupil_size_remove_eye_blinks(default_obj, buffer=10)
    assert np.sum(np.isnan(processed.pupil_size))==0
    processed = data.pupil_size_remove_eye_blinks(default_obj, buffer=100)
    assert np.sum(np.isnan(processed.pupil_size))==0

    processed = data.pupil_size_remove_eye_blinks(udef_obj, buffer=10)
    assert np.sum(np.isnan(processed.pupil_size))==0
    processed = data.pupil_size_remove_eye_blinks(udef_obj, buffer=100)
    assert np.sum(np.isnan(processed.pupil_size))==0

def test_shuffle():
    sess = default_obj.create_session()
    sess_u = udef_obj.create_session()
    cleaned_default = data.pupil_size_remove_eye_blinks(default_obj, buffer=50)
    event_timestamps = np.array(cleaned_default.trial_markers['start']) + 1000  # Buffer for baslining
    test_epochs = cleaned_default.create_epochs(event_timestamps,
                                                conditions=None,
                                                pre_event=200,
                                                post_event=200,
                                                pupil_baseline=[-200,-100])
    rand_shuf = session.shuffle(test_epochs)

    assert isinstance(rand_shuf, session.Epochs)
    for i in rand_shuf.data:
        assert i in test_epochs.data

    cleaned_udef = data.pupil_size_remove_eye_blinks(udef_obj, buffer=50)
    event_timestamps = np.array(cleaned_udef.trial_markers['start']) + 1000  # Buffer for baslining
    test_epochs_udef = cleaned_udef.create_epochs(event_timestamps,
                                                conditions=None,
                                                pre_event=200,
                                                post_event=200,
                                                pupil_baseline=[-200,-100])
    rand_shuf = session.shuffle(test_epochs_udef)
    assert isinstance(rand_shuf, session.Epochs)
    for i in rand_shuf.data:
        assert i in test_epochs_udef.data

    rand_sess = session.shuffle(sess)
    assert isinstance(rand_sess, session.Session)
    for i in rand_sess.data:
        assert i in sess.data


    rand_sess= session.shuffle(sess_u)
    assert isinstance(rand_sess, session.Session)
    for i in rand_sess.data:
        assert i in sess_u.data



import numpy as np



def linear_interpolate(data):
    """
    A function to linearly interpolate the data of a signal
    """
    nans = np.isnan(data)
    x = lambda z: z.nonzero()[0]
    data[nans] = np.interp(x(nans), x(~nans), data[~nans])
    return data
